/*  Screem:  screem-preferences.c
 *
 *  Copyright (C) 2002 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <config.h>

#include <string.h>

#include <glade/glade.h>

#include <gconf/gconf-client.h>

#include <libgnome/gnome-i18n.h>

#include <libgnomeui/gnome-font-picker.h>
#include <libgnomeui/gnome-entry.h>
#include <libgnomeui/gnome-file-entry.h>
#include <libgnomeui/gnome-icon-entry.h>
#include <libgnomeui/gnome-color-picker.h>

#include <gtk/gtkentry.h>
#include <gtk/gtktogglebutton.h>
#include <gtk/gtkspinbutton.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtkcombo.h>
#include <gtk/gtkliststore.h>
#include <gtk/gtktreeview.h>
#include <gtk/gtktreeselection.h>
#include <gtk/gtkcellrenderer.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkframe.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkpaned.h>
#include <gtk/gtknotebook.h>

#include <gdk/gdkcolor.h>
#include <gdk/gdkkeysyms.h>

#include "screem-application.h"
#include "screem-preferences.h"
#include "screem-helper.h"

#include "support.h"

#include "xml.h"

#include "screem-pref-dialog.h"

static void screem_preferences_dialog_response( GtkWidget *dialog, guint button,
						gpointer data );


typedef struct {
	const gchar *widget;
	const gchar *key;
	ScreemPrefNotifyType type;
	gint value;
} ScreemPrefStruct;

static ScreemPrefStruct misc_prefs[] = {
	{ "default_filename_entry", "/apps/screem/default_filename",       SCREEM_PREF_STRING,	0 },
	{ "default_charset_combo",  "/apps/screem/editor/default_charset", SCREEM_PREF_STRING,	0 },
	{ "default_dtd_combo",      "/apps/screem/editor/default_dtd",     SCREEM_PREF_STRING,	0 },
	{ NULL, NULL, 0, 0 }
};

static ScreemPrefStruct editor_colours[] = {
	{ "fontpicker",		"/apps/screem/editor/font",		SCREEM_PREF_STRING,	0 },
	{ "editortextpicker", 	"/apps/screem/editor/text", 		SCREEM_PREF_COLOUR,	0 },
	{ "editorbgpicker", 	"/apps/screem/editor/back", 		SCREEM_PREF_COLOUR,	0 },
	{ "themefont",		"/apps/screem/editor/themefont",	SCREEM_PREF_TOGGLE,	0 },
	{ "themecolours",	"/apps/screem/editor/themecolours",	SCREEM_PREF_TOGGLE,	0 },
	{ NULL, NULL, 0, 0 }
};

static ScreemPrefStruct editor_features[] = {
	{ "highlight", 		  "/apps/screem/editor/highlight", 	 SCREEM_PREF_TOGGLE,	0 },
	{ "tagging", 		  "/apps/screem/editor/inline_tagging",  SCREEM_PREF_TOGGLE,	0 },
	{ "indent", 		  "/apps/screem/editor/auto_indent", 	 SCREEM_PREF_TOGGLE,	0 },
	{ "tagclose", 		  "/apps/screem/editor/intelliclose",    SCREEM_PREF_TOGGLE,	0 },
	{ "wordwrap",		  "/apps/screem/editor/wordwrap",	 SCREEM_PREF_TOGGLE,	0 },
	{ "tabwidth", 	   	  "/apps/screem/editor/tabwidth",	 SCREEM_PREF_SPIN,	0 },
	{ "autoenton",		  "/apps/screem/editor/entityinsertion", SCREEM_PREF_RADIO,	0 },
	{ "autoentoff",		  "/apps/screem/editor/entityinsertion", SCREEM_PREF_RADIO,	1 },
	{ "autoentbychar",	  "/apps/screem/editor/entityinsertion", SCREEM_PREF_RADIO,	2 },
	{ NULL, NULL, 0, 0 }
};

static ScreemPrefStruct browser_prefs[] = {
	{ "browser1_static",  "/apps/screem/Browser1_static",  SCREEM_PREF_TOGGLE,	0 },
	{ "browser2_static",  "/apps/screem/Browser2_static",  SCREEM_PREF_TOGGLE,	0 },
	{ "browser3_static",  "/apps/screem/Browser3_static",  SCREEM_PREF_TOGGLE,	0 },
	{ "browser1_dynamic", "/apps/screem/Browser1_dynamic", SCREEM_PREF_TOGGLE,	0 },
	{ "browser2_dynamic", "/apps/screem/Browser2_dynamic", SCREEM_PREF_TOGGLE,	0 },
	{ "browser3_dynamic", "/apps/screem/Browser3_dynamic", SCREEM_PREF_TOGGLE,	0 },
	{ "browser1_path",    "/apps/screem/Browser1_path",    SCREEM_PREF_STRING,	0 },
	{ "browser2_path",    "/apps/screem/Browser2_path",    SCREEM_PREF_STRING,	0 },
	{ "browser3_path",    "/apps/screem/Browser3_path",    SCREEM_PREF_STRING,	0 },
	{ "browser1_icon",    "/apps/screem/Browser1_icon",    SCREEM_PREF_STRING,	0 },
	{ "browser2_icon",    "/apps/screem/Browser2_icon",    SCREEM_PREF_STRING,	0 },
	{ "browser3_icon",    "/apps/screem/Browser3_icon",    SCREEM_PREF_STRING,	0 },
	{ NULL, NULL, 0, 0 }
};

static ScreemPrefStruct print_prefs[] = {
	{ "print_headers", "/apps/screem/editor/print/headers",     SCREEM_PREF_TOGGLE,	0 },
	{ "wrap_lines",    "/apps/screem/editor/print/wrap_lines",  SCREEM_PREF_TOGGLE,	0 },
	{ "print_lines",   "/apps/screem/editor/print/lines",       SCREEM_PREF_TOGGLE,	0 },
	{ "lines", 	   "/apps/screem/editor/print/nlines",	    SCREEM_PREF_SPIN,	0 },
	{ NULL, NULL, 0, 0 }
};


static const gchar *charsets[] = {
	"ISO-8859-1",
	"ISO-8859-2",
	"ISO-8859-5",
	"ISO-8859-7",
	"UTF-8",
	NULL
};

static void screem_preferences_edit_macros( ScreemApplication *application,
					    GladeXML *xml );
static void screem_preferences_edit_helpers( ScreemApplication *application,
					     GladeXML *xml );
static void screem_preferences_helper_select( GtkTreeSelection *sel );

static void screem_preferences_macro_select( GtkTreeSelection *sel );


static void screem_preferences_add_page( GtkWidget *dialog,
					GladeXML *xml, const gchar *root,
					ScreemPrefStruct *pref,
					const gchar *cat, const gchar *name,
					const gchar *stock )
{
	guint i;
	GtkWidget *widget;
	
	for( i = 0; pref[ i ].widget; ++ i ) {
		widget = glade_xml_get_widget( xml, pref[ i ].widget );
		screem_pref_dialog_add_pref( SCREEM_PREF_DIALOG( dialog ),
						pref[ i ].key,
						pref[ i ].type,
						pref[ i ].value,
						widget );
	}
	widget = glade_xml_get_widget( xml, root );
	screem_pref_dialog_add_page_with_stock( SCREEM_PREF_DIALOG( dialog ),
						cat, name, stock, widget );
	g_object_ref( widget );

}

void screem_preferences_edit( ScreemApplication *application )
{
	GConfClient *client;
	gchar *gladepath;
	GladeXML *xml;
	GtkWidget *widget;
	gint i;
	static GtkWidget *dialog = NULL;
	GList *list;
	GdkPixbuf *logo;
	
	if( dialog ) {
		gtk_widget_show( dialog );
		gdk_window_raise( GDK_WINDOW( dialog->window ) );
		return;
	}

	logo = gdk_pixbuf_new_from_file( SPLASH_DIR"/splash.png", NULL );
	dialog = GTK_WIDGET( screem_pref_dialog_new_with_logo( logo ) );
	g_object_unref( logo );

	client = gconf_client_get_default();
	
	/* create dialog */
	gladepath = gconf_client_get_string( client, 
					     "/apps/screem/general/glade_path",
					     NULL );
	if( ! gladepath ) {
		gladepath = g_strconcat( GLADE_PATH, "/screem.glade",
					 NULL );
	}

	/**** editor colour prefs ****/
	xml = glade_xml_new( gladepath, "editor_colours", NULL );
	glade_xml_signal_autoconnect( xml );	
	screem_preferences_add_page( dialog,
				xml, "editor_colours", editor_colours, 
				_( "Editor" ), _( "Fonts & Colours" ),	
				GTK_STOCK_SELECT_COLOR );
	g_object_unref( xml );
	

	/**** editor feature prefs ****/
	xml = glade_xml_new( gladepath, "editor_features", NULL );
	glade_xml_signal_autoconnect( xml );	
	screem_preferences_add_page( dialog,
				xml, "editor_features", editor_features, 
				_( "Editor" ), _( "Features" ),	
				GTK_STOCK_PREFERENCES );
	g_object_unref( xml );
	
	/****  misc prefs ****/
	xml = glade_xml_new( gladepath, "misc_prefs", NULL );

	widget = glade_xml_get_widget( xml, "default_dtd_combo" );
	screem_dtd_fill_combo( application, widget );
	
	widget = glade_xml_get_widget( xml, "default_charset_combo" );
	for( list = NULL, i = 0; charsets[ i ]; ++ i ) {
		list = g_list_prepend( list, (gpointer)charsets[ i ] );
	}
	list = g_list_reverse( list );
	gtk_combo_set_popdown_strings( GTK_COMBO( widget ), list );
	g_list_free( list );
	glade_xml_signal_autoconnect( xml );
	screem_preferences_add_page( dialog,
				xml, "misc_prefs", misc_prefs, 
				_( "General" ), _( "Defaults" ),	
				GTK_STOCK_PREFERENCES );
	g_object_unref( xml );
	
	/****  external browsers ****/
	xml = glade_xml_new( gladepath, "external_browsers", NULL );
	glade_xml_signal_autoconnect( xml );
	screem_preferences_add_page( dialog,
				xml, "external_browsers", browser_prefs,
				_( "Browsers" ), _( "External" ),
				"Screem_Preview" );
	g_object_unref( xml );

	
	/**** editor macros ****/
	xml = glade_xml_new( gladepath, "macro_prefs", NULL );
	screem_preferences_edit_macros( application, xml );
	glade_xml_signal_autoconnect( xml );
	widget = glade_xml_get_widget( xml, "macro_prefs" );
	screem_pref_dialog_add_page_with_stock( SCREEM_PREF_DIALOG( dialog ),
						_( "Editor" ), _( "Macros" ),
						GTK_STOCK_PREFERENCES, widget );
	g_object_ref( widget );
	g_object_set_data( G_OBJECT( dialog ), "macrosxml", xml );

	/**** helpers ****/
	xml = glade_xml_new( gladepath, "helper_prefs", NULL );
	screem_preferences_edit_helpers( application, xml );
	glade_xml_signal_autoconnect( xml );
	widget = glade_xml_get_widget( xml, "helper_prefs" );
	screem_pref_dialog_add_page_with_stock( SCREEM_PREF_DIALOG( dialog ),
						_( "General" ), _( "Helpers" ),
						GTK_STOCK_EXECUTE, widget );
	g_object_ref( widget );
	g_object_set_data( G_OBJECT( dialog ), "helpersxml", xml );

	/* printing */
	xml = glade_xml_new( gladepath, "print_prefs", NULL );
	glade_xml_signal_autoconnect( xml );
	screem_preferences_add_page( dialog,
				xml, "print_prefs", print_prefs,
				_( "Editor" ), _( "Print" ),
				GTK_STOCK_PRINT );
	g_object_unref( xml );

	
	g_free( gladepath );

	g_signal_connect( G_OBJECT( dialog ), "response",
			  G_CALLBACK( screem_preferences_dialog_response ),
			  &dialog );

	gtk_widget_show( dialog );

	g_object_unref( client );
}

static void screem_preferences_dialog_response( GtkWidget *dialog, guint button,
						gpointer data )
{
	GtkWidget **dialogptr;
	GladeXML *xml;
	
	dialogptr = data;
	*dialogptr = NULL;

	xml = g_object_get_data( G_OBJECT( dialog ), "macrosxml" );
	g_object_unref( xml );
	xml = g_object_get_data( G_OBJECT( dialog ), "helpersxml" );
	g_object_unref( xml );
	
	gtk_widget_destroy( dialog );
}

static void screem_preferences_edit_macros( ScreemApplication *application,
					    GladeXML *xml )
{
	GtkListStore *store;
	GList *macros;

	GtkWidget *view;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *col;
	const gchar *titles[] = { "Key Combo", "Action/Text", NULL };
	gint i;
	GtkTreeSelection *selection;

	store = gtk_list_store_new( 3, G_TYPE_STRING, G_TYPE_STRING,
				    G_TYPE_POINTER, NULL );
	/* build model */
	for( macros = application->macros; macros; macros = macros->next ) {
		GtkTreeIter it;
		ScreemEditorMacro *macro;

		macro = (ScreemEditorMacro*)macros->data;

		gtk_list_store_append( store, &it );
		gtk_list_store_set( store, &it, 
				    0, macro->key_name,
				    1, macro->text,
				    2, macro, -1 );
	}

	/* create tree view */
	view = glade_xml_get_widget( xml, "macrolist" );
	g_object_set_data( G_OBJECT( view ) , "application", application );

	for( i = 0; titles[ i ]; ++ i ) {
		renderer = gtk_cell_renderer_text_new();
		col = gtk_tree_view_column_new();
		gtk_tree_view_column_set_title( col, titles[ i ] );
		gtk_tree_view_column_pack_start( col, renderer, TRUE );
		gtk_tree_view_append_column( GTK_TREE_VIEW( view ), col );
		gtk_tree_view_column_set_attributes( col, renderer, 
						     "text", i, NULL );
	}
	
	gtk_tree_view_set_model( GTK_TREE_VIEW( view ), 
				 GTK_TREE_MODEL( store ) );
	g_object_unref( store );

	selection = gtk_tree_view_get_selection( GTK_TREE_VIEW( view ) );
	g_signal_connect( G_OBJECT( selection ), "changed",
			  G_CALLBACK( screem_preferences_macro_select ),
			  NULL );
}

/* also used for the update button as we perform a check to see
   if a helper already exists */
void screem_preferences_add_macro( GtkWidget *widget )
{
	GladeXML *xml;
	GtkWidget *view;
	ScreemEditorMacro *macro;
	ScreemApplication *application;
	GtkTreeModel *model;
	GtkTreeIter it;

	const gchar *combo;
	const gchar *action;

	xml = glade_get_widget_tree( widget );

	view = glade_xml_get_widget( xml, "macrolist" );
	model = gtk_tree_view_get_model( GTK_TREE_VIEW( view ) );

	application = SCREEM_APPLICATION( g_object_get_data( G_OBJECT( view ),
							     "application" ) );

	widget = glade_xml_get_widget( xml, "macrocombo" );
	combo = gtk_entry_get_text( GTK_ENTRY( widget ) );

	widget = glade_xml_get_widget( xml, "macroaction" );
	widget = GTK_COMBO( widget )->entry;
	action = gtk_entry_get_text( GTK_ENTRY( widget ) );

	macro = g_new0( ScreemEditorMacro, 1 );
	macro->key_name = g_strdup( combo );
	macro->text = g_strdup( action );

	screem_application_add_macro( application, macro );

	gtk_list_store_append( GTK_LIST_STORE( model ), &it );
	gtk_list_store_set( GTK_LIST_STORE( model ), &it, 
			    0, macro->key_name,
			    1, macro->text,
			    2, macro, -1 );
}

void screem_preferences_remove_macro( GtkWidget *widget )
{
	GladeXML *xml;
	GtkWidget *view;
	ScreemEditorMacro *macro;
	ScreemApplication *application;
	GtkTreeModel *model;
	GtkTreeIter it;
	GtkTreeSelection *selection;

	const gchar *combo;

	xml = glade_get_widget_tree( widget );

	view = glade_xml_get_widget( xml, "macrolist" );
	selection = gtk_tree_view_get_selection( GTK_TREE_VIEW( view ) );

	application = SCREEM_APPLICATION( g_object_get_data( G_OBJECT( view ),
							     "application" ) );

	if( gtk_tree_selection_get_selected( selection, &model, &it ) ) {

		widget = glade_xml_get_widget( xml, "macrocombo" );
		combo = gtk_entry_get_text( GTK_ENTRY( widget ) );
		
		macro = g_new0( ScreemEditorMacro, 1 );
		macro->key_name = (gchar*)combo;
		screem_application_remove_macro( application, macro );
		g_free( macro );

		gtk_list_store_remove( GTK_LIST_STORE( model ), &it );
	}
}

static void screem_preferences_macro_select( GtkTreeSelection *sel )
{
	GtkTreeView *view;
	GtkTreeModel *model;
	GtkTreeIter it;
	GtkWidget *widget;
	GladeXML *xml;

    	view = gtk_tree_selection_get_tree_view( sel );

	xml = glade_get_widget_tree( GTK_WIDGET( view ) );

	if( gtk_tree_selection_get_selected( sel, &model, &it ) ) {
		GValue nvalue = {0};
		GValue pvalue = {0};
		GtkWidget *widget;

		gtk_tree_model_get_value( model, &it, 0, &nvalue );
		gtk_tree_model_get_value( model, &it, 1, &pvalue );
		
		widget = glade_xml_get_widget( xml, "macrocombo" );
		gtk_entry_set_text( GTK_ENTRY( widget ),
				    g_value_get_string( &nvalue ) );

		widget = glade_xml_get_widget( xml, "macroaction" );
		widget = GTK_COMBO( widget )->entry;
		gtk_entry_set_text( GTK_ENTRY( widget ),
				    g_value_get_string( &pvalue ) );

		g_value_unset( &nvalue );
		g_value_unset( &pvalue );
	} else {
		widget = glade_xml_get_widget( xml, "macrocombo" );
		gtk_entry_set_text( GTK_ENTRY( widget ), "" );
		widget = glade_xml_get_widget( xml, "macroaction" );
		widget = GTK_COMBO( widget )->entry;
		gtk_entry_set_text( GTK_ENTRY( widget ), "" );
	}
}

gboolean screem_preferences_grab_key( GtkWidget *widget, GdkEventKey *event )
{
	gboolean ret;
	gchar *key;

	ret = FALSE;

	if( event->keyval != GDK_Tab &&
	    event->keyval != GDK_Up &&
	    event->keyval != GDK_Right &&
	    event->keyval != GDK_Down &&
	    event->keyval != GDK_Left ) {
		key = convert_keysym_state_to_string (event->keyval,
						      event->state);
		if( key ) {
			gtk_entry_set_text (GTK_ENTRY(widget), key );
			g_free( key );
		} else {
			gtk_entry_set_text (GTK_ENTRY(widget), "" );
		}
		ret = TRUE;
	}

	return ret;
}

static void screem_preferences_edit_helpers( ScreemApplication *application,
					     GladeXML *xml )
{
	GtkListStore *store;
	GList *helpers;
	GtkWidget *view;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *col;
	GtkTreeSelection *selection;
	const gchar *titles[] = { "Name", "Path", "Icon", NULL };
	gint i;


	store = gtk_list_store_new( 4, 
				    G_TYPE_STRING,
				    G_TYPE_STRING, G_TYPE_STRING,
				    G_TYPE_POINTER,
				    NULL );
	/* build model */
	for(helpers = application->helpers; helpers; helpers = helpers->next) {
		GtkTreeIter it;
		ScreemHelper *helper;
		const gchar *name;
		const gchar *pathname;
		const gchar *icon;

		helper = SCREEM_HELPER( helpers->data );

		g_object_get( G_OBJECT( helper ), 
			      "name", &name, "pathname", &pathname,
			      "iconpath", &icon, NULL );
		
		gtk_list_store_append( store, &it );

		gtk_list_store_set( store, &it,
				    0, name, 
				    1, pathname,
				    2, icon,
				    3, helper,
				    -1 );
	}

	/* create tree view */
	view = glade_xml_get_widget( xml, "helperlist" );
	g_object_set_data( G_OBJECT( view ) , "application", application );

	for( i = 0; titles[ i ]; ++ i ) {
		renderer = gtk_cell_renderer_text_new();
		col = gtk_tree_view_column_new();
		gtk_tree_view_column_set_title( col, titles[ i ] );
		gtk_tree_view_column_pack_start( col, renderer, TRUE );
		gtk_tree_view_append_column( GTK_TREE_VIEW( view ), col );
		gtk_tree_view_column_set_attributes( col, renderer, 
						     "text", i, NULL );
	}
	
	gtk_tree_view_set_model( GTK_TREE_VIEW( view ), 
				 GTK_TREE_MODEL( store ) );
	g_object_unref( store );

	selection = gtk_tree_view_get_selection( GTK_TREE_VIEW( view ) );
	g_signal_connect( G_OBJECT( selection ), "changed",
			  G_CALLBACK( screem_preferences_helper_select ),
			  NULL );
}

/* also used for the update button as we perform a check to see
   if a helper already exists */
void screem_preferences_add_helper( GtkWidget *widget )
{
	GladeXML *xml;
	GtkWidget *view;
	ScreemHelper *helper;
	ScreemApplication *application;
	GtkTreeModel *model;
	GtkTreeIter it;

	const gchar *name;
	const gchar *pathname;
	gchar *iconpath;

	HelperInputMode imode;
	HelperOutputMode omode;
	
	gboolean found;
	ScreemHelper *old;
	gchar *oname;
	
	xml = glade_get_widget_tree( widget );

	view = glade_xml_get_widget( xml, "helperlist" );
	model = gtk_tree_view_get_model( GTK_TREE_VIEW( view ) );

	application = SCREEM_APPLICATION( g_object_get_data( G_OBJECT( view ),
							     "application" ) );

	widget = glade_xml_get_widget( xml, "helpername" );
	widget = gnome_entry_gtk_entry( GNOME_ENTRY( widget ) );
	name = gtk_entry_get_text( GTK_ENTRY( widget ) );

	widget = glade_xml_get_widget( xml, "helperpath" );
	widget = gnome_file_entry_gtk_entry( GNOME_FILE_ENTRY( widget ) );
	pathname = gtk_entry_get_text( GTK_ENTRY( widget ) );

	widget = glade_xml_get_widget( xml, "helpericon" );
	iconpath = gnome_icon_entry_get_filename( GNOME_ICON_ENTRY( widget ) );

	imode = SCREEM_HELPER_STDIN_NONE;
	widget = glade_xml_get_widget( xml, "stdin_page" );
	if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( widget ) ) ) {
		imode = SCREEM_HELPER_STDIN_SELECTION;
	}
	widget = glade_xml_get_widget( xml, "stdin_user" );
	if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( widget ) ) ) {
		imode = SCREEM_HELPER_STDIN_USER;
	}
	
	omode = SCREEM_HELPER_STDOUT_NONE;
	widget = glade_xml_get_widget( xml, "stdout_page" );
	if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( widget ) ) ) {
		omode = SCREEM_HELPER_STDOUT_SELECTION;
	}
	widget = glade_xml_get_widget( xml, "stdout_insert" );
	if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( widget ) ) ) {
		omode = SCREEM_HELPER_STDOUT_INSERT;
	}
	
	helper = screem_helper_new( name, pathname, iconpath,
				    imode, omode );

	screem_application_remove_helper( application, helper );
	screem_application_add_helper( application, helper );

	found = FALSE;
	old = NULL;
	if( gtk_tree_model_get_iter_first( GTK_TREE_MODEL( model ),
					&it ) ) {
		do {
			old = NULL;
			oname = NULL;
			gtk_tree_model_get( GTK_TREE_MODEL( model ),
					&it, 3, &old, -1 );
			if( old ) {
				g_object_get( G_OBJECT( old ),
					"name", &oname,
					NULL );
			}
			found = ( oname && ! strcmp( name, oname ) );
		} while( ! found &&
			gtk_tree_model_iter_next( GTK_TREE_MODEL( model ),
						&it ) );
	}

	if( ! found ) {
		old = NULL;
		gtk_list_store_append( GTK_LIST_STORE( model ), &it );
	}
	gtk_list_store_set( GTK_LIST_STORE( model ), &it,
			    0, name, 
			    1, pathname,
			    2, iconpath,
			    3, helper,
			    -1 );

	g_free( iconpath );

	if( old ) {
		g_object_unref( old );
	}
}

void screem_preferences_remove_helper( GtkWidget *widget )
{
	GladeXML *xml;
	GtkWidget *view;
	ScreemHelper *helper;
	ScreemApplication *application;
	GtkTreeModel *model;
	GtkTreeIter it;
	GtkTreeSelection *selection;

	xml = glade_get_widget_tree( widget );

	view = glade_xml_get_widget( xml, "helperlist" );
	selection = gtk_tree_view_get_selection( GTK_TREE_VIEW( view ) );

	application = SCREEM_APPLICATION( g_object_get_data( G_OBJECT( view ),
							     "application" ) );
	
	if( gtk_tree_selection_get_selected( selection, &model, &it ) ) {
		gtk_tree_model_get( GTK_TREE_MODEL( model ),
				&it, 3, &helper, -1 );
	
		screem_application_remove_helper( application, helper );

		gtk_list_store_remove( GTK_LIST_STORE( model ), &it );

		g_object_unref( G_OBJECT( helper ) );
	}
}

static void screem_preferences_helper_select( GtkTreeSelection *sel )
{
	GtkTreeView *view;
	GtkTreeModel *model;
	GtkTreeIter it;
	GtkWidget *widget;
	GladeXML *xml;

       	view = gtk_tree_selection_get_tree_view( sel );

	xml = glade_get_widget_tree( GTK_WIDGET( view ) );

	if( gtk_tree_selection_get_selected( sel, &model, &it ) ) {
		GValue nvalue = { 0 };
		GValue pvalue = { 0 };
		GValue ivalue = { 0 };
		GValue hvalue = { 0 };
		GladeXML *xml;
		GtkWidget *widget;
		ScreemHelper *helper;
		gint imode;
		gint omode;
		const gchar *iwidgets[] = {
			"stdin_none", "stdin_page", "stdin_user"
		};
		const gchar *owidgets[] = {
			"stdout_msg", "stdout_page", "stdout_insert"
		};
		
		gtk_tree_model_get_value( model, &it, 0, &nvalue );
		gtk_tree_model_get_value( model, &it, 1, &pvalue );
		gtk_tree_model_get_value( model, &it, 2, &ivalue );
		gtk_tree_model_get_value( model, &it, 3, &hvalue );
		
		xml = glade_get_widget_tree( GTK_WIDGET( view ) );
		
		widget = glade_xml_get_widget( xml, "helpername" );
		widget = gnome_entry_gtk_entry( GNOME_ENTRY( widget ) );
		gtk_entry_set_text( GTK_ENTRY( widget ),
				    g_value_get_string( &nvalue ) );

		widget = glade_xml_get_widget( xml, "helperpath" );
		widget = gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(widget));
		gtk_entry_set_text( GTK_ENTRY( widget ),
				    g_value_get_string( &pvalue ) );

		widget = glade_xml_get_widget( xml, "helpericon" );
		gnome_icon_entry_set_filename( GNOME_ICON_ENTRY( widget ),
					       g_value_get_string( &ivalue ) );

		helper = g_value_get_pointer( &hvalue );
		g_assert( helper );
		g_object_get( G_OBJECT( helper ), 
				"input", &imode, 
				"output", &omode, NULL );
		widget = glade_xml_get_widget( xml, iwidgets[ imode ] );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( widget ),
						TRUE );
		widget = glade_xml_get_widget( xml, owidgets[ omode ] );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( widget ),
						TRUE );

		widget = glade_xml_get_widget( xml, "helperremove" );
		gtk_widget_set_sensitive( widget, TRUE );

		g_value_unset( &nvalue );
		g_value_unset( &pvalue );
		g_value_unset( &ivalue );
		g_value_unset( &hvalue );
	} else {
		widget = glade_xml_get_widget( xml, "helpername" );
		widget = gnome_entry_gtk_entry( GNOME_ENTRY( widget ) );
		gtk_entry_set_text( GTK_ENTRY( widget ), "" );

		widget = glade_xml_get_widget( xml, "helperpath" );
		widget = gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(widget));
		gtk_entry_set_text( GTK_ENTRY( widget ), "" );

		widget = glade_xml_get_widget( xml, "helpericon" );
		gnome_icon_entry_set_filename( GNOME_ICON_ENTRY( widget ), "");
		widget = glade_xml_get_widget( xml, "helperremove" );
		gtk_widget_set_sensitive( widget, FALSE );
	}
}


